/*
  Copyright (c) 2008, Arturas Moskvinas. All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License. You should have
  received a copy of the GPL license along with this program; if you
  did not, you can find it at http://www.gnu.org/
*/

#ifndef _GNU_SOURCE 
#define _GNU_SOURCE
#endif
#include "httpd.h"
#include "http_log.h"
#include "http_request.h"
#include "ap_mpm.h"
#include "scoreboard.h"
/*getpid*/
#include <sys/types.h>
#include <unistd.h>
/*kill and killpg*/
#include <signal.h>
/*apr_atoi64 and apr_int64_t*/
#include "apr.h"
#include "apr_strings.h"
#include "unixd.h"
module AP_MODULE_DECLARE_DATA timedout_module;
/*apr < 1.3 compatiblity issue*/
#ifndef APR_FPROT_UWRITE
#define APR_FPROT_UWRITE APR_UWRITE
#endif
static int server_limit = 0;
static int thread_limit = 0;

/*configuration*/
typedef struct timedout_cfg {
    apr_int64_t enabled;
    apr_int64_t kill_one;
    apr_int64_t log_one;
    apr_int64_t only_warn;
    apr_int64_t leave_process;
    apr_int64_t leave_longest_running_one;
    apr_int64_t min_free_servers;
    apr_int64_t long_request_time;
    apr_int64_t log_requests;
    apr_int64_t kill_signal;
    char *log_lock_path;
    char *kill_lock_path;
} timedout_cfg;

/*process PID and running time*/
typedef struct pid_rtime {
    apr_time_t rtime;
    pid_t pid;
} pid_rtime;

static void* timedout_dconf_create(apr_pool_t *p, char * dir)
{
    timedout_cfg *cfg;
    cfg = (timedout_cfg *) apr_palloc(p, sizeof(timedout_cfg));
    cfg->enabled = -1; /*not enabled by default*/
    cfg->kill_one = -1;
    cfg->log_one = -1;
    cfg->only_warn = -1;
    cfg->leave_process = -1;
    cfg->leave_longest_running_one = -1;
    cfg->min_free_servers = -1;
    cfg->long_request_time = -1;
    cfg->log_requests = -1;
    cfg->log_lock_path = "/tmp/mod_timedout_log.lock";
    cfg->kill_lock_path = "/tmp/mod_timedout_kill.lock";
    cfg->kill_signal = 9;
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &thread_limit);
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS, &server_limit);
    return (void *)cfg;
}

/*values are inherited from parent if it is not set in current context*/
/* not needed, because I timedout is not able to figure out to which virtual host (or especially dir/location) request belongs...
static void *timedout_dconf_merge(apr_pool_t *p, void *conf1, void *conf2)
{
    
    timedout_cfg *cfg;
    cfg = (timedout_cfg *) apr_palloc(p, sizeof(timedout_cfg));
    timedout_cfg *over_conf1 = (timedout_cfg *) conf1;
    timedout_cfg *over_conf2 = (timedout_cfg *) conf2; 

    cfg->enabled = over_conf2->enabled!=-1 ? over_conf2->enabled : over_conf1->enabled;

    return (void *)cfg;
}*/

static const char* setenableflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->enabled = arg1;
    return NULL;
}

static const char* setsurvivortype(cmd_parms *cmd, void *mconfig, const int arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->leave_longest_running_one = arg1;
    return NULL;
}


static const char* setlongrequesttime(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->long_request_time = apr_atoi64(arg1);
    return NULL;
}

static const char* setkillsurvivoramount(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->leave_process = apr_atoi64(arg1);
    return NULL;
}

static const char* setminfreeservers(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->min_free_servers = apr_atoi64(arg1);
    return NULL;
}

static const char* setkillsignal(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->kill_signal = apr_atoi64(arg1);
    return NULL;
}


static const char* setwarnonly(cmd_parms *cmd, void *mconfig, const int arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->only_warn = arg1;
    return NULL;
}
static const char* setonekillflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->kill_one = arg1;
    return NULL;
}

static const char* setonelogflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->log_one = arg1;
    return NULL;
}


static const char* setlogrequests(cmd_parms *cmd, void *mconfig, const int arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->log_requests = arg1;
    return NULL;
}


static const char* setonekilllockpath(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->kill_lock_path = (char *)arg1;
    return NULL;
}


static const char* setoneloglockpath(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    timedout_cfg *cfg = mconfig;
    cfg->log_lock_path = (char *)arg1;
    return NULL;
}


/*Log currently running process */
static void log_requests(request_rec *r)
{
    worker_score *w_record;
    int i, j;
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, "mod_timedout: Currently running requests:");
    for (i=0; i < server_limit; i++) /*daemons*/
    {
        for (j=0; j < thread_limit; j++)/*threads*/
        {
    	    w_record = ap_get_scoreboard_worker(i,j);
	    if ((w_record->status==SERVER_READY) || (w_record->status==SERVER_DEAD)) /*not running workers*/
	    {
	        continue;
	    }
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, "mod_timedout[%li]: Hostname:%s, uri: %s, client: %s, pid: %i", r->connection->id, w_record->vhost, w_record->request, w_record->client, ap_get_scoreboard_process(i)->pid);
	}
    }
}

static pid_rtime* collect_long_run_proc(apr_pool_t *pool, apr_int64_t max_request_time, int *elem_count)
{
    int elem, i, j;
    pid_rtime *p_time_array;
    elem = 0;
    p_time_array = apr_pcalloc(pool, sizeof(pid_rtime)*server_limit*thread_limit);
    worker_score *w_record;

    for (i=0; i < server_limit; i++) /*daemons*/
    {
        for (j=0; j < thread_limit; j++)/*threads*/
        {/*not ready for server with multiple threads... TODO*/
	    w_record = ap_get_scoreboard_worker(i,j);
	    switch (w_record->status)
	    {
		case SERVER_STARTING:
		case SERVER_READY:
		case SERVER_DEAD:
		case SERVER_IDLE_KILL:
		    break;
		default:
		    if (apr_time_sec(apr_time_now()-w_record->start_time)>max_request_time)
		    {
			p_time_array[elem].pid=ap_get_scoreboard_process(i)->pid;
			p_time_array[elem++].rtime+=apr_time_sec(apr_time_now()-w_record->start_time);
		    }
	    }
	}
    }
    *elem_count = elem;
    return p_time_array;
}

static int pid_rtime_order(const void *a, const void *b)
{
    return ((pid_rtime *)b)->rtime - ((pid_rtime *)a)->rtime;
}

static void kill_long_running_requests(request_rec *r, apr_int64_t kill_signal, apr_int64_t max_request_time, apr_int64_t leave_process, apr_int64_t leave_longest_running_one, apr_int64_t only_warn)
{
    int i, elem_count, start_inc, stop_dec;
    pid_rtime *p_time_array;

    p_time_array=collect_long_run_proc(r->pool, max_request_time, &elem_count);
    if (elem_count>0)
    {
	start_inc = 0;
	stop_dec = 0;
	qsort(p_time_array, elem_count+1, sizeof(pid_rtime), pid_rtime_order);
	if (leave_longest_running_one>0)
	{
	    start_inc = leave_process>0?leave_process:0;
	}
	else
	{
	    stop_dec = leave_process>0?-leave_process:0;
	}
	for (i=start_inc; i<elem_count+stop_dec; i++)
	{
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, "mod_timedout: Killing long running process, pid %i, runtime %li", p_time_array[i].pid, p_time_array[i].rtime);
	    if (only_warn<1)
	    {
		if (getpid()!=p_time_array[i].pid)
		{
		    kill(p_time_array[i].pid, kill_signal);
		}
		else
		{
		    ap_log_error(APLOG_MARK, APLOG_CRIT, 0, r->server, "mod_timedout: Tried to kill itself!");
		}
	    }
	}
    }
}

static apr_status_t lock_file(apr_file_t **file, const char *path, apr_pool_t *pool, server_rec *s)
{
    apr_status_t status;
    char buffer[256];
    if (NULL==path)
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_timedout: %s", "Bad path argument passed");
	return APR_BADARG;
    }
    status = apr_file_open(file, path, APR_CREATE | APR_WRITE , APR_FPROT_UWRITE, pool);
    if (APR_SUCCESS!=status)
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_timedout: File open failed, reason %s", apr_strerror(status, buffer, 255));	
	return status;
    }
    status = apr_file_lock(*file, APR_FLOCK_EXCLUSIVE | APR_FLOCK_NONBLOCK);
    if (APR_SUCCESS!=status)
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_timedout: File lock failed, reason %s", apr_strerror(status, buffer, 255));
	apr_file_close(*file);
    }
    return status;
}

static apr_status_t unlock_file(apr_file_t *file, server_rec *s)
{
    apr_status_t status;
    char buffer[256];
    status = apr_file_unlock(file);
    if (status!=APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_timedout: Failed to unlock file, reason %s", apr_strerror(status, buffer, 255));	
    }
    return apr_file_close(file);
}

static int count_free_servers()
{
    int i, j, sum;
    worker_score *w_record;
    sum = 0;/*how many free servers*/
    for (i=0; i < server_limit; i++) /*daemons*/
    {
        for (j=0; j < thread_limit; j++)/*threads*/
        {
	    w_record = ap_get_scoreboard_worker(i,j);
	    /*we do not care if process is gracefully being restarted, starting up or other state, because module should startup when server limit is being reached*/
	    /*so we count only ready and dead ones*/
	    /*dead process will be spawned if needed*/
	    if ((w_record->status==SERVER_READY)|| (w_record->status==SERVER_DEAD))
	    {
		sum++;
	    }
	}
    }
    return sum;
}

static int count_long_requests(apr_int64_t long_request_time)
{
    int i, j, sum;
    worker_score *w_record;
    sum = 0;/*how many free servers*/
    for (i=0; i < server_limit; i++) /*daemons*/
    {
        for (j=0; j < thread_limit; j++)/*threads*/
        {
	    w_record = ap_get_scoreboard_worker(i,j);
	    /*we do not care if process is gracefully being restarted, starting up or other state, because module should startup when server limit is being reached*/
	    /*so we count only ready and dead ones*/
	    /*dead process will be spawned if needed*/
	    switch (w_record->status)
	    {
		case SERVER_STARTING:
		case SERVER_READY:
		case SERVER_DEAD:
		case SERVER_IDLE_KILL:
		    break;
		default:
		    if (apr_time_sec(apr_time_now()-w_record->start_time)>long_request_time)
		    {
			sum++;
		    }
	    }
	}
    }
    return sum;
}


static int mod_timedout_method_handler (request_rec *r)
{
    server_rec *s;
    timedout_cfg *cfg;
    s = r->server;
    apr_file_t *file;
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_timedout: %s", "Handling request");
    cfg = (timedout_cfg*) ap_get_module_config(r->per_dir_config, &timedout_module);

    /*no config or not enabled*/
    if ((cfg==NULL) || (cfg->enabled==0))
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_timedout: %s", "Doing nothing, timedout not enabled");
	return DECLINED;
    }

    if (cfg->min_free_servers<0)
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,"mod_timedout: %s", "Min free server count is not set, bailing out");
	return DECLINED;
    }
    
    if (cfg->leave_process<0)
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,"mod_timedout: %s", "Long running process count which is being left after detection of timeout is not set, bailing out");
	return DECLINED;
    }

    if (cfg->leave_longest_running_one<0)
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,"mod_timedout: %s", "TimedOutLeaveLongestRunningRequestKill is not set, bailing out");
	return DECLINED;
    }

    if (cfg->long_request_time<0)
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,"mod_timedout: %s", "Long running request time is not set, bailing out");
	return DECLINED;
    }

    if (cfg->log_requests<0)
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,"mod_timedout: %s", "Request logging is not set (please set to log or not to log), bailing out");
	return DECLINED;
    }

    /*TODO: extra check for some configuration values, to disambigue everything*/
    if (count_free_servers()<cfg->min_free_servers)
    {	/*OK limit reached, lets start checking for long requests*/
	/*check if there are more than minimum allowed long running request*/
	ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_timedout: %s, current count %i, the mimimum allowed %li", "Free server count is low, start checking",count_free_servers(), cfg->min_free_servers);
	if (count_long_requests(cfg->long_request_time)>cfg->leave_process)
	{/*ok we need to clean up*/
	    /*first log request */
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_timedout: there are to many long running request, start cleaning up, there are %i long running requests, allowed %li", count_long_requests(cfg->long_request_time), cfg->leave_process);
	    if ((cfg->log_requests) && ((cfg->log_one<=0) || (lock_file(&file, cfg->log_lock_path, r->pool, r->server)==APR_SUCCESS)))
	    {
		log_requests(r);
		if (cfg->log_one>0)
		{
		    unlock_file(file, r->server);
		}
	    }
	    /*now kill requests*/
	    if (cfg->kill_one<=0 || (lock_file(&file, cfg->kill_lock_path, r->pool, r->server)==APR_SUCCESS))
	    {
		/*customize kill signal*/
		kill_long_running_requests(r, 9, cfg->long_request_time, cfg->leave_process, cfg->leave_longest_running_one, cfg->only_warn);
		if (cfg->kill_one>0)
		{
		    unlock_file(file, r->server);
		}
	    }
	}
    }
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_timedout: %s", "Not TimedOuted, returning OK");
    return DECLINED;
}

static void mod_timedout_register_hooks (apr_pool_t *p)
{
    ap_hook_handler(mod_timedout_method_handler, NULL, NULL, APR_HOOK_FIRST);
}

static const command_rec timedout_cmds[] = {

AP_INIT_FLAG(
"TimedOutEnable",
setenableflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Enable TimedOut module"),

AP_INIT_TAKE1(
"TimedOutLongRunningRequestTime",
setlongrequesttime,
NULL,
RSRC_CONF|ACCESS_CONF,
"Time needed to become long running request"),

AP_INIT_TAKE1(
"TimedOutLeaveLongRunningRequest",
setkillsurvivoramount,
NULL,
RSRC_CONF|ACCESS_CONF,
"How many long running request to keep running (eg. to finish some cron or to allow DB to cache some request which are waited by those requests)"),

AP_INIT_FLAG(
"TimedOutLeaveLongestRunningRequest",
setsurvivortype,
NULL,
RSRC_CONF|ACCESS_CONF,
"Flag to enable/disable feature which will kill/warn and leave the longest running long process or vice versa"),


AP_INIT_TAKE1(
"TimedOutMinFreeServers",
setminfreeservers,
NULL,
RSRC_CONF|ACCESS_CONF,
"If there are only count or less free server, then start search 'n kill"),

AP_INIT_FLAG(
"TimedOutWarnOnly",
setwarnonly,
NULL,
RSRC_CONF|ACCESS_CONF,
"Request are not killed, only warned that it would be killed, for example to check if settings are not to strict"),

AP_INIT_FLAG(
"TimedOutLogOneAtATime",
setonelogflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Module will log one request list at a time, to avoid congestion"),

AP_INIT_FLAG(
"TimedOutLogRequests",
setlogrequests,
NULL,
RSRC_CONF|ACCESS_CONF,
"Log request when timedout conditions are met"),

AP_INIT_FLAG(
"TimedOutKillOneAtATime",
setonekillflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Module will start one killing process at a time, to avoid congestion"),

AP_INIT_TAKE1(
"TimedOutLogLockFile",
setoneloglockpath,
NULL,
RSRC_CONF|ACCESS_CONF,
"Lock file needed to implement one log at a time"),

AP_INIT_TAKE1(
"TimedOutKillLockFile",
setonekilllockpath,
NULL,
RSRC_CONF|ACCESS_CONF,
"Lock file needed to implement one kill at a time"),

AP_INIT_TAKE1(
"TimedOutKillSignal",
setkillsignal,
NULL,
RSRC_CONF|ACCESS_CONF,
"Signal used to kill apache processes, default 9 (SIGKILL)"),


{NULL}
};


/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Data declarations.                                                       */
/*                                                                          */
/* Here are the static cells and structure declarations private to our      */
/* module.                                                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/
module AP_MODULE_DECLARE_DATA timedout_module =
{
    STANDARD20_MODULE_STUFF,
    timedout_dconf_create,    /* per-directory config creator */
    NULL,     /* rules are the same as per dir... */
    NULL,/*timedout_dconf_create, */                    /* server config creator */
    NULL, /*timedout_dconf_merge, */                    /* server config merger */
    timedout_cmds,              	/* command table */
    mod_timedout_register_hooks,       /* callback for registering hooks */
};
